import java.awt.Rectangle;

//int irisX; //relative to image origin
//int irisY; //relative to image origin
//int diameter;

class EyeTracker {
  public int tL; //tolerance for left
  public int tR; //.. right
  public int tU; //.. up
  public int tD; //.. down

  boolean calibrating;
  int calibratingStage;

  float irisX_0 = 0; //average x-Coord of eye; /relative to image origin
  float irisY_0 = 0; //"" y; relative to image origin
  int irisX; //iris position, relative to image origin, result of each hough-transform
  int irisY;
  int irisDiameter; 

  public Rectangle feature;

  private int imageWidth;
  private int imageHeight;
  private int imageX; //coordinates for top-left corner of live camera image in the screen
  private int imageY; 

  private int contourWhiteReflectionThreshold = 120;
  private int originalBlackPupilThreshold = 80;

  public PGraphics HoughOutputPlane;
  private PGraphics output;
  public static final int a = 0;

  color bgcolor;


  private FeatureDetection fd;
  private GData gg;
  CamContainer cc;

  PApplet parent;



  EyeTracker(PApplet parent, CamContainer cc) {
    this.imageWidth = cc.cam.width;
    this.imageHeight = cc.cam.height;
    this.imageX = cc.imgCenterX;
    this.imageY = cc.imgCenterY;

    this.parent = parent;
    this.cc = cc;
    this.gg = new GData();
    this.fd = new FeatureDetection(parent, EYE, this.cc.w, this.cc.h);
    this.fd.setPosition(cc.imgCenterX, cc.imgCenterY);
    this.HoughOutputPlane = createGraphics(this.imageWidth, this.imageHeight);
    this.output = createGraphics(imageWidth, imageHeight);
    println(imageWidth, imageHeight);
    this.calibrating = true;
    this.calibratingStage = 1;
  }








  void calibrate() { //main calibration method
    switch(this.calibratingStage) {
    case 0: 
      calibrateCam(); //first calibrate light of cam
      break;
    case 1: 
      calibrateFD(); //then look for eyes/faces
      break;
    case 2: 
      calibrateHough(); //ajust eye-diameter size
      break;
    case 3: 
      calibrateEyePosition();
      break;
    default: 
      endCalibration();
      break;
    }
  }






  void calibrateCam() { //0
    background(bgcolor);

    output.beginDraw();
    output.set(this.imageX, this.imageY, this.cc.cam);
    output.endDraw();
    image(output, 200, 200);

    fill(max(bgColor-20, 0), 120);
    rect(0, 0, 500, 150);
    fill(textColor);
    textSize(40);
    text("Calibrating Camera", 20, 50);
    textSize(20);
    text("Press enter to resume", 20, 90);
  }






  void calibrateFD() { //calibratingState = 1
    background(bgcolor);

    output.beginDraw();
    output.set(0, 0, this.cc.cam);
    Rectangle[] noses = fd.detectNose(cc);
    for (int i = 0; i < noses.length; i++) {
      rect(noses[0], output, false, 180);
    }
    output.endDraw();
    image(output, imageX, imageY);

    this.fd.detect(true, this.cc, this.calibratingStage);

    fill(max(bgColor-20, 0), 120);
    rect(0, 0, 500, 150);
    text("Searching for Eyes...", 40, 20, 50, textColor);
    text("Scroll to change detection tolerance", 20, 20, 90, textColor);
    text("Select one Eye...", 20, 20, 120, textColor);
    rect(width-40, this.fd.getTolerance(), 40, 25);
    textSize(14);
    text(this.fd.getTolerance(), width-60, this.fd.getTolerance()+20);
  }






  void calibrateHough() { //calibratingState = 2
    background(bgcolor);
    output.beginDraw();
    gg.setImg(cc.cam);
    gAlgorithms();
    output.set(0, 0, this.gg.img);
    output.rect(this.feature.x+cc.imgCenterX, this.feature.y+cc.imgCenterY, this.feature.width, this.feature.height);
    output.endDraw();

    hough(this.feature.x, this.feature.y, this.feature.width+this.feature.x, this.feature.height+this.feature.y, (int)this.gg.param2, (int)this.gg.param2, 2);

    image(output, imageX, imageY);
    image(HoughOutputPlane, imageX, imageY);
  }








  //find average eye position, when looked to center of the screen; needs several runs!
  //__:
  int caliTime0 = 1; //time to calibrate in seconds
  int caliTime = caliTime0*generalFrameRate; //current calibrate countdwon number

  float averageFeatureX = 0; //average center x-Coord of eye-rectangle (openCV), is ...relativ... to image origin
  float averageFeatureY = 0; //"" center y, is ...relative... to image origin
  float averageFeatureCountAddedValues = 0; //count how many values are being added to "averageFeatureX/Y" (Sometimes openCV doesn't detect any eyes).
  int c = 0;
  int rel = 1;
  int calitime2 = caliTime0*generalFrameRate*rel;


  void calibrateEyePosition2() { //calibratingStage = 3
    if (calibratingStage == 3) {
      if (calitime2 == caliTime0*generalFrameRate*rel) {
        frameRate(rel*generalFrameRate);
        delay(1000);
        CalibrationTimer ct = new CalibrationTimer(this, caliTime0*generalFrameRate);
        ct.start();
        println("AASHDGJHASGDJASGDJHASGDJASGDJAGSUDZATGSDHAGSDJAHSGD");
      }

      //background(bgColor);
      println(bgColor);
      /*
      //text("Look to \nthe center of the screen!", 50, 0, 40, textColor);
       colorMode(RGB, 255);
       fill(100, 100, 100, 50);
       float d = (caliTime0 * rel * generalFrameRate - calitime2) / (float)(caliTime0 * rel * generalFrameRate) * width;
       //ellipse(width/2, height/2, d, d);
       fill(224, 216, 0, 180);*/
      //ellipse(width/2, height/2, 100, 100);
      calitime2--;
    }
  }



  void calibrateEyePosition() { //calibratingState = 3
    if (caliTime0 * generalFrameRate == caliTime) {
      background(bgColor);
      irisDiameter = gg.param2;

      text("Look to \nthe center of the screen!", 50, 0, 40, textColor);
      textSize(20);
      fill(200, 0, 0);
      ellipse(width/2, height/2, 100, 100);
    }

    if (caliTime0 * generalFrameRate - 1 == caliTime) {
      delay(1000);
    }

    if (caliTime > 0) {
      background(bgColor);
      output.beginDraw();
      output.clear();

      Rectangle[] noses = fd.detectNose(cc);
      if (noses.length > 0) {
        Rectangle currentFeature = noses[0];//fitRightRectangle(noses, feature, fittingTolerance, threshold);
        if (currentFeature != null) {
          drawRectNoFill(noses[0], output);
          averageFeatureX += getCenterOfRect(noses[0]).x;
          averageFeatureY += getCenterOfRect(noses[0]).y;
          averageFeatureCountAddedValues++;
        }
      }

      this.gg.setImg(cc.cam);
      gAlgorithms();
      set(cc.imgCenterX, cc.imgCenterY, gg.img);
      hough(feature.x, feature.y, feature.width + feature.x, feature.height + feature.y, (int)gg.param2, (int)gg.param2, 2);

      output.endDraw();
      image(output, imageX, imageY);
      irisX_0 += irisX;
      irisY_0 += irisY;

      colorMode(RGB, 255);
      text("Look to the center of the screen!", 50, 0, 40, textColor);
      fill(100, 100, 100, 50);
      float d = (caliTime0 * generalFrameRate - caliTime) / (float)(caliTime0 * generalFrameRate)*width;
      ellipse(width/2, height/2, d, d);
      //fill(200, 0, 0);
      fill(224, 216, 0, 180);
      ellipse(width/2, height/2, 100, 100);
      caliTime--;
    } else { 
      //finish 
      irisX_0 /= caliTime0 * generalFrameRate;
      irisY_0 /= caliTime0 * generalFrameRate;
      averageFeatureX /= averageFeatureCountAddedValues;
      averageFeatureY /= averageFeatureCountAddedValues;
      relX = irisX_0 - averageFeatureX;
      relY = irisY_0 - averageFeatureY;

      text(averageFeatureX, 60, 60);
      println("AVERGEFEAUTREX=", averageFeatureX);
      calibratingStage++;
    }
  }






  float relX, relY; //distance between iris center and nose corner
  Rectangle nose;

  Rectangle fitRightRectangle(Rectangle[] array, Rectangle target, float fittingTolerance, int threshold) {
    int compareValue[] = new int[array.length+2]; //array to store the "quality" of a feature; determined by how close it is to the originally selected feature
    for (int i = 0; i < array.length; i++) { 
      if (abs(target.x - array[i].x) < fittingTolerance)
        compareValue[i]++;
      if (abs(target.y - array[i].y) < fittingTolerance)
        compareValue[i]++;
      if (abs(target.width - array[i].width) < fittingTolerance)
        compareValue[i]++;
      if (abs(target.height - array[i].height) < fittingTolerance)
        compareValue[i]++;
    }
    //get highest value
    int maxCompareValue = 0;
    int maxCompareValueIndex = -1;
    for (int i = 0; i < compareValue.length; i++) {
      if (compareValue[i] > maxCompareValue) {
        maxCompareValue = compareValue[i];
        maxCompareValueIndex = i;
      }
    }
    if (maxCompareValue > threshold)
      return array[maxCompareValueIndex];
    else return null;
  }

  void endCalibration() {
    calibrating = false;
  }

  void resetCalib() {
    irisX_0 = 0;
    irisY_0 = 0;
    averageFeatureX = 0; //average x-Coord of eye-rectangle (openCV), is ...relativ... to image origin
    averageFeatureY = 0; //"" y, is ...relative... to image origin
    averageFeatureCountAddedValues = 0;
    calibratingStage = 1;
    calibrating = true;
    caliTime = caliTime0*generalFrameRate;
    calitime2 = caliTime0*generalFrameRate*rel;
  }










  void gAlgorithms() {

    //myG.simpleBlur();
    //myG.lowPass(myG.param2);
    int t0 = millis();
    this.gg.blur();
    //this.gg.advancedContour(feature.x-50, feature.y-50, feature.x + feature.width+50, feature.y+feature.height+50);
    this.gg.advancedContour(0,0,gg.w, gg.h);
    //myG.lowPass(myG.param2);
    //myG.threshold(myG.param1);
    //myG.optimize();
    println("gg-algos-time:", millis()-t0);
  }








  public void track() {
    background(bgColor);
    output.beginDraw();
    output.clear();

    if (liveVideo && !liveOutput) {
      set(0, 0, cc.cam);
    }

    //copy image from webcam to grahpic-editing object 
    this.gg.setImg(cc.cam);


    //apply graphic algorithms
    gAlgorithms();
    if (liveOutput) {
      output.set(0, 0, this.gg.img);
    }

    //search for circles
    hough(feature.x, feature.y, feature.width + feature.x, feature.height + feature.y, irisDiameter, irisDiameter, 2);

    //evaluate result coordinates (finalX, finalY) of hough-transform 
    evaluate();

    //get nose position
    Rectangle[] noses = fd.detectNose(cc); 
    if (noses.length > 0) {
      irisX_0 = getCenterOfRect(noses[0]).x + relX;
      irisY_0 = getCenterOfRect(noses[0]).y + relY;
    }

    //if (detectFeaturesLive && ++observeDetectionFrequency % (generalFrameRate/faceDetectionFrequency) == 0 || true) { }
    if (liveOutput) {
      if (noses.length > 0)
        drawRectNoFill(noses[0], output);
      output.endDraw();
      image(output, imageX, imageY);
    }
  }







  public void setToleranceL(int t) {
    if (t >= 0) {
      this.tL = t;
    }
  }
  public void setToleranceR(int t) {
    if (t >= 0) {
      this.tR = t;
    }
  }
  public void setToleranceU(int t) {
    if (t >= 0) {
      this.tU = t;
    }
  }
  public void setToleranceD(int t) {
    if (t >= 0) {
      this.tD = t;
    }
  }









  void hough(int x0, int y0, int xMax, int yMax, int rMin, int rMax, int rasterResolution) { //d:circle diameter, r:raster resolution
    //intializing
    //int t0 = millis(); //for time performance measurement, get starting time
    xMax = floor((xMax-x0)/rasterResolution)*rasterResolution + x0; //set xMax to a multiple of rasterResolution
    output.beginDraw();
    output.noFill();
    output.strokeWeight(1);
    output.stroke(200);
    if (irisX_0 != 0) {
      output.ellipse(irisX_0, irisY_0, irisDiameter, irisDiameter);
      println(999);
    }

    int numberHits = 1;                    //determines how many maxima will be searched

    IntList cValues = new IntList();       //contains the results of the hough-operation
    int cMaxs[] = new int[numberHits];     //array for the 5 highest values
    int cMaxsPos[] = new int[numberHits];  //array for the position if the same 5 highest values
    IntList posX = new IntList();          //contains x position of value
    IntList posY = new IntList();          //contains y position of value
    IntList rx = new IntList();            //contains radius of value
    int t0 = millis();
    int wid = cc.cam.width;
    //EVENTUELL SPÄTER NUR NACH HALBKREISEN SUCHEN
    for (int r = rMin; r <= rMax; r++) {
      for (int y = y0; y < yMax; y += rasterResolution) {
        for (int x = x0; x < xMax; x += rasterResolution) {
          if (brightness(gg.img.pixels[y * wid + x]) > contourWhiteReflectionThreshold || brightness(gg.img2.pixels[y * gg.img.width + x]) < originalBlackPupilThreshold) { //überprüft punkt nur wenn Mittelpunkt des Kreises weiß ist. Funktioniert nur unter der Prämisse, dass die Pupille teilweise weiß ist (Reflexion des Bildschirms)
            int value = 0;
            for (int i = 0; i < 360; i += 5) {
              int xx = (int)(r/2*cos(radians(i)))+x;
              int yy = (int)(r/2*sin(radians(i)))+y;
              if (yy*wid+xx >= 0 && yy*wid+xx+2 < gg.img.pixels.length) {
                value += brightness(gg.img.pixels[yy*wid+xx]);
                value += brightness(gg.img.pixels[yy*wid+xx+2]); //make radius quasi-wider (here only in x-direction), so thicker circles can be better recognized
              }
            }
            cValues.append(value);
            posX.append(x);
            posY.append(y);
            rx.append(r);
          }
        }
      }
    }
    println("HOUGH: " + (millis()-t0));

    //get max values
    for (int k = 0; k < min(cMaxs.length, cValues.size()); k++) {
      cMaxs[k] = 0;
      for (int j = 0; j < cValues.size(); j++) {
        if (cValues.get(j) > cMaxs[k]) {
          cMaxs[k] = cValues.get(j);
          cMaxsPos[k] = j;
        }
      }
      cValues.set(cMaxsPos[k], 0);
    }

    output.colorMode(HSB, 255);
    output.stroke(200, 200, 200);
    int clrCircle = 255; //circle-color, so highest accuracies are marked more red, lower rather blue
    irisDiameter = rMin;

    for (int k = 0; k < min(numberHits, cValues.size()); k++) {
      HoughOutputPlane.stroke(clrCircle, 200, 200);
      clrCircle -=50;
      output.ellipse(posX.get(cMaxsPos[k]), posY.get(cMaxsPos[k]), irisDiameter, irisDiameter);
      irisDiameter = rx.get(cMaxsPos[k]);
    }
    if (posX.size() > 0) {
      //println("Hough - "+cValues.size()+" values, max: "+cMaxs[0] +"("+posX.get(cMaxsPos[0]), posY.get(cMaxsPos[0])+"), r: " +diameter+", t=" + (millis()-t0));
      irisX = posX.get(cMaxsPos[0]);
      irisY = posY.get(cMaxsPos[0]);
    } else { 
      //println("no values");
    }
    output.endDraw();
  }
}



class CalibrationTimer extends Thread {
  private boolean active;
  private EyeTracker parent;
  private int numberMeasurements;


  CalibrationTimer(EyeTracker parent, int numberMeasurements) {
    this.parent = parent;
    active = false;
    this.numberMeasurements = numberMeasurements;
  }


  public void start() {
    if (!active) {
      active = true;
      super.start();
    }
  }


  void run() {
    while (true) {
      if (active) {
        execute();
      } else {
        break;
      }
    }
  }


  private void execute() {
    float avgNoseX = 0;
    float avgNoseY = 0;
    float avgIrisX = 0;
    float avgIrisY = 0;
    EyeTracker e = new EyeTracker(this.parent.parent, cc);

    for (int i = 0; i < numberMeasurements; i++) {
      Rectangle[] noses = e.fd.detectNose(cc);

      if (noses.length > 0) {
        Rectangle currentFeature = noses[0];//fitRightRectangle(noses, feature, fittingTolerance, threshold);
        if (currentFeature != null) {
          //drawRectNoFill(noses[0], output);
          avgNoseX += getCenterOfRect(noses[0]).x;
          avgNoseY += getCenterOfRect(noses[0]).y;
        }
      }

      //parent.gg.setImg(cc.cam);
      //parent.gAlgorithms();
      //set(cc.imgCenterX, cc.imgCenterY, gg.img);
      //parent.hough(feature.x, feature.y, feature.width + feature.x, feature.height + feature.y, (int)parent.gg.param2, (int)parent.gg.param2, 2);

      //output.endDraw();
      //image(output, imageX, imageY);
      avgIrisX += e.irisX;
      avgIrisY += e.irisY;
    }
    avgIrisX /= numberMeasurements;
    avgIrisY /= numberMeasurements;
    avgNoseX /= numberMeasurements;
    avgNoseY /= numberMeasurements;
    parent.relX = avgIrisX - avgNoseX;
    parent.relY = avgIrisY - avgNoseY;

    parent.calibratingStage++;
    quit();
  }


  public void quit() {
    active = false;
    interrupt();
  }
}
