
/*
  This sketch deals with the detection and processing of the eyes and camera data. 
 */



import java.awt.Rectangle; 



/*
 * SETTINGS
 */
int generalFrameRate = 20;

boolean detectFeaturesLive = true;
float faceDetectionFrequency = 10; //times per second
int observeDetectionFrequency = 0;
boolean displayRects = true;

boolean editGraphic = true;
boolean applyHough = true;
boolean useCam = true;
int camNumber = 25;
//int camNumber = 40;
boolean liveVideo = true;
boolean liveOutput = true;
boolean motionDetection = false; //turns liveVideo to false automatically
boolean fastCam = false;

//Interface
color bgColor = 20;
color textColor = 180;
PImage loading;
int loadSize;
int loadHeight;

CamContainer cc;
EyeTracker et;


PImage img; //used for motion detection, last image is saved
Rectangle feature = new Rectangle(50, 50, 100, 100); //contains data for the position of the selected eye



void setup() {
  fullScreen();
  frameRate(60);                     //will be set later, because high fps will be needed for starting animation
  cc = new CamContainer();           //no initializing yet, still a list of available cameras is being fetched in another thread already
  cc.printCameraList();              //print available cameras
  cc.selectCamera(camNumber);        //select camera index from list
  loading = loadImage("loading-symbol.png");
  loadSize = width/8;


  if (useCam) {
    cc.initCam(this);
  }
  background(bgColor);
  /*if (motionDetection) { liveVideo = false; try {this.img = (PImage)cc.cam.clone();} catch(Exception e){ e.printStackTrace();} } */
}


int rotateAngle = 0;
/*__________________D R A W____________________________*/

void draw() {
  if (cc.isInitialized()) { //cam needs to be initialized, aka gottenn the first image, so image dimensions can be passed to featureDetection
    if (et == null) {
      et = new EyeTracker(this, cc);
      cc.printCameraList();
      rotate(0);
      textAlign(LEFT);
      frameRate(generalFrameRate);
    } else {
      if (et.calibrating) {
        et.calibrate();     //run through the calibration process
      } else {
        et.track();         //main function, runs most of the time

        text(toleranceU, 1 /2. * (width-100)+50, 0 / 2. *(height-100)+50);
        text(toleranceD, 1 /2. * (width-100)+50, 2 / 2. *(height-100)+50);
        text(toleranceL, 2 /2. * (width-100)+50, 1 / 2. *(height-100)+50);
        text(toleranceR, 0 /2. * (width-100)+50, 1 / 2. *(height-100)+50);
      }
    }
  } else {
    background(bgColor);
    cc.initCam(this);      //checks if camera list is ready and then starts initializing
    textAlign(CENTER);
    textSize(40);
    text("Searching for cameras...", width/2, height/5);
    
    translate(width/2, height/2);
    rotate(rotateAngle*TWO_PI/360);
    translate(-loadSize/2, -loadSize/2);
    
    image(loading, 0, 0, loadSize, loadSize);
    
    rotateAngle += 2;
  }
}

/*_______________E N D D R A W_________________________*/

int toleranceL = 6;
int toleranceR = 6;
int toleranceU = 6;
int toleranceD = 6;

int measure = 1;
int lastDir;
int formerDirX = 1;
int formerDirY = 1;

void evaluate() {
  int dirX = 1;
  int dirY = 1;

  colorMode(RGB, 255);
  fill(0, 150, 200);
  strokeWeight(0);
  if (et.irisX - (et.irisX_0) > toleranceL) { //left
    dirX = 0;
  }
  if (et.irisX - (et.irisX_0) < -toleranceR) { //right
    dirX = 2;
  }
  if (et.irisY - (et.irisY_0) < -toleranceU) { //up
    dirY = 0;
  }
  if (et.irisY - (et.irisY_0) > toleranceD) { //down
    dirY = 2;
  }
  int dir = 3*dirY + dirX;
  //println(dir);
  if (measure > 1) {
    formerDirX = dirX;
    formerDirY = dirY;
  }
  rect(formerDirX / 2. * (width-100), formerDirY / 2. *(height-100), 100, 100);
  if (lastDir == dir) {
    measure++;
  } else {
    measure = 0;
  }
  lastDir = dir;
}





/*void evaluateOld1() {
 print("\n");
 colorMode(RGB, 255);
 fill(0, 150, 200);
 strokeWeight(0);
 if (finalX - (averageX) > toleranceX) {
 print("left"); 
 rect(0, height/2-50, 100, 100);
 }
 if (finalX - (averageX) < -toleranceX) {
 print("right");
 rect(width-100, height/2-50, 100, 100);
 }
 if (finalY - (averageY) < -toleranceY) {
 print("up");
 rect(width/2-50, 0, 100, 100);
 }
 if (finalY - (averageY) > toleranceY) {
 print("down");
 rect(width/2-50, height-100, 100, 100);
 }
 print("\n");
 }*/






void motionDetection(int t) {
  int t0 = millis();
  for (int i = 0; i < img.pixels.length; i++) {
    float b = brightness(img.pixels[i])-brightness(cc.cam.pixels[i]);
    if (b > t)
      img.pixels[i] = color(0, 0, 255);
    else 
    img.pixels[i] = color(0, 0, 0);
  }
  //p1.pixels[i] = color(0, 0, min(255, 255*((brightness(p1.pixels[i])-brightness(p2.pixels[i])/t))));
  image(img, 0, 0);
  println(millis()-t0);
  try {
    this.img = (PImage)cc.cam.clone();
  }
  catch(Exception e) {
    e.printStackTrace();
  }
}
