#include "FlexSensor.h";

int onLedPin = 3;
int sleepLedPin = 5;


FlexSensor myFlexSensor(7, 5);    //initialize with pin 7 and threshold 5


void setup() {
  Serial.begin(9600);
  myFlexSensor.setLedPins(onLedPin, sleepLedPin); //set pin numbers to set lights indicating the state of the arduino
}


void loop() {
  if (myFlexSensor.isActive()) {        //only run code, when program is "active"

    if (!myFlexSensor.isCalibrated()) { //if not yet calibrate then do it

      myFlexSensor.calibrate();

    } else {

      myFlexSensor.readData();
      delay(100);

    }
  }
}



String input;


 /*
  * By sending data through the USB port the arduino can set various states and settings without having to upload a new program. 
  * A command has to be followed by the newline character (\n). 
  * 
  * The Arduino can change some states by receiving a single-digit number (followed by \n):
  *   "2" - Set the Arduino to sleep, it won't really stop the arduino (which is impossible) but it will no further send any messages nor measure to state of the flex-sensor. 
  *   "1" - Wake the Arduino again. 
  *   "3" - Recalibrate the sensor. 
  *   "4" - Enable distinguishing the direction in which the flex-sensor is bent.
  *   "5" - Disable it. 
  *   "6" - Enable long click. 
  *   "7" - Disable long click, every signal will be evaluated with "short"
  *   "8" - The Arduino will start to send the measured voltage in 10bit relative to the calibrated value and stop to evaluate the data. Useful when testing for errors
  *   "9" - Disable it again. 
  *   
  * 
  * Moreover thetolerance can be set via computer by writing the letter t followed by the new tolerance value: i.e. sending "t24\n" would set the tolerance to +-24. 
  * Same goes for the thresholds for distinguishing short (s) and long (l) signals.
  * 
  * 
  */



void serialEvent() {

  char inputChar;
  int inputInt;

  while (Serial.available()) {
    inputChar = Serial.read();
    input += inputChar;

    if (inputChar == '\n') {
      if (input.toInt()) {
        switch (input.toInt()) {
          case 1: myFlexSensor.setActive(1); break;  //set arduino working state to active
          case 2: myFlexSensor.setActive(0); break;  //set arduino working state to inactive
          case 3: myFlexSensor.reset(); break;//call a reset on variables. also recalibrate
          case 4: myFlexSensor.enableBothDirections(1); break;
          case 5: myFlexSensor.enableBothDirections(0); break;
          case 6: myFlexSensor.enableLongClick(1); break;
          case 7: myFlexSensor.enableLongClick(0); break;
          case 8: myFlexSensor.enableContinuousValues(1); break;
          case 9: myFlexSensor.enableContinuousValues(0); break;
          default: Serial.println("error");
        }
      } else if (input.startsWith("t")) { //set tolerance
        if (input.substring(1).toInt() > -1) {
          myFlexSensor.setTolerance(input.substring(1).toInt());
        } else {
          Serial.println("error w/t");
        }
      } else if ( input.startsWith("s")) { //set short thresh
        if (input.substring(1).toInt() > -1) {
          myFlexSensor.setShortThresh(input.substring(1).toInt());
        } else {
          Serial.println("error w/sT");
        }
      } else if ( input.startsWith("l")) { //set long thresh
        if (input.substring(1).toInt() > -1) {
          myFlexSensor.setLongThresh(input.substring(1).toInt());
        } else {
          Serial.println("error w/lT");
        }
      } else {
        Serial.println("synerror");
      }

      input = "";
    }
  }
}
