/*
 *
 * The MultilineTextBox is part of the gui library for processing created by us (see the tab Z_GUI).
 * 
 * Like the name says it's a textbox that registers keystrokes when having been clicked on and inserts the keys. 
 * When reaching the end of a line it automatically breaks the line and scrolls down if the total height of the text exceeds the objects height. 
 *
 * Don't expect to be able select any text or set the cursor with the mouse! (; (everything's done manually here)
 *    (arrow keys work though, so do backspace and del)
 *
 *
 *
 *
 */

public class MultilineTextBox extends VScrollContainer {

  /*
   *  STYLE
   */

  protected String content = "";     //text content

  protected StringList lines;        //text in lines
  protected IntList lineBreaks;      //marks the automatic and entered line-breaks



  protected int Cursor = TEXT;
  public int paddingLeft;
  public int paddingRight;
  public int paddingTop;
  public int paddingBottom;



  protected int cursorPosition = 0;

  //not to set by the user
  protected int lineHeight;
  protected boolean autoScroll; //autoscroll to cursor when text changed



  protected int cursorTime;
  protected boolean displayCurs = true;


  MultilineTextBox() {
    lines = new StringList();
    lineBreaks = new IntList();
    stickyFocus = true;
    ForegroundColor = 0;
    lineHeight = FontSize + 10;
    addMouseListener("");
    //cursor blink animation:
    cursorTime = millis();
    registerMethod("pre", this);
    paddingLeft = 3;
    paddingLeft = 4;
    paddingTop = 3;
    paddingBottom = 3;
  }


  /*
   * Graphics
   */


  @Override
    public void render() {
    int t0 = millis();

    //first get total virtual height of all text lines 

    pg = createGraphics(Width + 5, Height); //add some extra space for the scroll bar
    pg.beginDraw();
    pg.textSize(FontSize);
    pg.textAlign(LEFT, TOP);
    pg.fill(ForegroundColor);
    if (VisualBackgroundColor != 0) {
      pg.background(VisualBackgroundColor);
    }
    pg.stroke(70);
    pg.line(0, 0, Width + 5, 0);
    pg.line(0, 0, 0, Height);

    boxedText(content, X, Y, Width);

    fullScrollHeight = lines.size() * lineHeight;
    //DRAW CURSOR
    if (this.Focused && displayCurs) {
      drawCursor();
    }

    //DRAW TEXT
    for (int i = 0; i < lines.size(); i++) {
      pg.text(lines.get(i), paddingLeft, i * lineHeight + paddingTop - scrollPositionY);
    }

    //DRAW SCROLLBAR
    float scrollbarHeight = (float)Height / fullScrollHeight * Height;
    if (scrollbarHeight < Height) { //don't display scroll-bar when there's nothing to scroll

      float scrollBarY = constrain(scrollPositionY * (float)(Height - scrollbarHeight) / (float)(fullScrollHeight - Height), 1, Height - scrollbarHeight - 2);
      pg.fill(255);
      pg.noStroke();
      pg.rect(Width + 1, scrollBarY, 3, scrollbarHeight, 15);
    }

    pg.endDraw();
    //println("TIME", millis()-t0);
  }










  void boxedText(String str, int x0, int y0, int w) {
    lines = new StringList();
    lineBreaks = new IntList();

    lineHeight = FontSize * 3 / 2;

    float spaceWidth = pg.textWidth(" ");
    float usedLineSpace = 0;
    String currentLine = "";
    int countChars = 0;

    lineBreaks.append(1);


    String paragraphs[] = split(str, '\n');

    for (int j = 0; j < paragraphs.length; j++) {
      String words[] = split(paragraphs[j], ' ');

      for (int i = 0; i < words.length; i++) {
        float wordWidth = pg.textWidth(words[i]);

        if (usedLineSpace /*+ spaceWidth*/ + wordWidth < w) { //no new line

          currentLine += words[i] + " ";
          usedLineSpace += spaceWidth + wordWidth;
          countChars += 1 + words[i].length();
        } else { //break line

          countChars += 1;
          lineBreaks.append(countChars);     //store position of break
          lines.append(currentLine);        //store previous line

          currentLine = words[i] + " ";       //new word is first word of currentLine
          usedLineSpace = wordWidth + spaceWidth;
          countChars += words[i].length();
        }
      }
      currentLine += "\n"; 
      //countChars += 1;
      lineBreaks.append(countChars);   
      lines.append(currentLine); 

      currentLine = "";
      usedLineSpace = 0;
    }
    lineBreaks.append(countChars);
    //lines.append(currentLine);
  }




  /*
   * CURSOR
   */


  //cursor blink animation
  void pre() {
    if (this.Focused) {
      int t = millis();
      if (t - cursorTime > 500) {
        displayCurs = !displayCurs;
        cursorTime = t;
        update();
      }
    }
  }

  void drawCursor() {
    int index = -1;
    for (int i = 0; i < lineBreaks.size(); i++) {
      if (cursorPosition < lineBreaks.get(i)) {
        index = i;
        //println(i);
        break;
      }
    }
    //println("cursor:", cursorPosition, lineBreaks);
    int cursorHeight = FontSize;
    if (index >= 0 && cursorPosition > 0) {
      //println("found linebreak", lineBreaks.get(index));
      int start = lineBreaks.get(index-1)-1;
      int stop = cursorPosition;//-lineBreaks.get(index-1)+1;
      //println("substring(" + start +", "+stop);
      String a = content.substring(start, stop);
      float wordWidth = pg.textWidth(a) + .5;

      int cursorY = (index-1) * lineHeight; //position of upper left corner of cursor relative to first character of text

      //autoscroll
      if (autoScroll) {
        if (cursorY - scrollPositionY + cursorHeight > Height - paddingBottom) {   //cursor left visible box at the bottom
          setScrollPositionY(cursorY - Height + FontSize + paddingTop + paddingBottom);
        } else if (cursorY < scrollPositionY + paddingTop) {          //cursor left visible box at the top
          setScrollPositionY(cursorY);
          println(8);
        }
        autoScroll = false;
      }

      //draw cursor
      pg.line(paddingLeft + wordWidth, cursorY - scrollPositionY + paddingTop, wordWidth + paddingLeft, cursorY + cursorHeight - scrollPositionY + paddingTop);
    } else {

      pg.line(paddingLeft, paddingTop, paddingLeft, cursorHeight + paddingTop);
    }
  }







  /*
   * EDIT TEXT
   */

  void append(char c) {
    cursorPosition += 1;
    content = content.substring(0, cursorPosition-1) + c + content.substring(cursorPosition-1);

    update();
    textChanged();
  }

  void append(String s) {
    cursorPosition += s.length();
    content = content.substring(0, cursorPosition-1) + s + content.substring(cursorPosition-1);

    update();
    textChanged();
  }

  void setContent(String newContent) {
    this.content = newContent;
    update();
    textChanged();
  }

  void backspace() {
    if (content.length() > 0) {

      content = content.substring(0, cursorPosition-1) + content.substring(cursorPosition);
      cursorPosition--;
      update();
      textChanged();
    }
  }

  void delete() {
    if (content.length() > cursorPosition) {

      content = content.substring(0, cursorPosition) + content.substring(cursorPosition+1);
      update();
      textChanged();
    }
  }

  void textChanged() {
    cursorTime = millis();
    displayCurs = true;
    autoScroll = true;
  }

  void moveCursor(int i) {
    cursorPosition = constrain(cursorPosition + i, 0, content.length());
    update();
    textChanged();
  }



  /*
   * EVENTS
   */

  @Override
    public void hover(boolean hovers, boolean isNewEvent) {
    if (isNewEvent) {
      if (hovers) {

        cursor(Cursor);
        update();
      } else {

        cursor(0);
        update();
      }
      update();
    }
  }

  @Override
    public void pressed(boolean pressed, boolean isNewEvent) {

    if (isNewEvent) {
      if (pressed) { //pressed

        if (!actionPerformedMethodName.equals("")) {
          method(actionPerformedMethodName);
        }
        this.Focus();
        cursorTime = millis();
        displayCurs = true;
        update();
      } else { //released
      }
    }
  }


  @Override 
    public void onKeyPress(KeyEvent e) {
    char KEY = e.getKey();
    int Code = e.getKeyCode();
    if (e.getKey() == CODED) {
      switch(Code) {
      case LEFT:
        moveCursor(-1);
        break;
      case RIGHT: 
        moveCursor(1);
        break;
      }
    } else { //not coded
      switch(KEY) {
      case BACKSPACE:
        backspace();
        break;
      case DELETE:
        delete();
        break;
      default: 
        append(KEY);
      }
      //println(KEY);
    }
  }
}








public class TextView extends Control {

  //no border
  public int BorderWidth = 0;
  public boolean Transparent = true;

  String word = "";                             //current word 


  int cursor = 0;                               //position of cursor in paragraph




  TextView(int Width) {
    super();
    Name = "LABEL";
    VisualBackgroundColor = color(0, 0);
    pg.beginDraw();
    this.Width = Width;
    Height = 60;
    pg = createGraphics(Width, Height);
  }



  /*
   *append at cursor
   */
  void append(String s) {
    cursor += s.length();
    if (s == " ") {
      word = "";
      cursor = 0;
    } else {
      word = word.substring(0, cursor-1) + s + word.substring(cursor-1);
    }
    update();
  }
  //@override 
  public void render() {
    pg.beginDraw();
    pg.clear();

    if (Transparent) {
      pg.noFill();
    } else {
      pg.fill(VisualBackgroundColor);
    }
    pg.noStroke();
    pg.rect(0, 0, Width, Height);

    pg.textSize(FontSize);
    pg.fill(ForegroundColor);
    pg.textAlign(LEFT, CENTER);
    pg.text(word, 0, Height/2);
    drawCursor();
    pg.endDraw();
  }




  void drawCursor() {
    float wordWidth = pg.textWidth(word.substring(0, cursor))+2;
    pg.stroke(0);
    pg.line(wordWidth, 10, wordWidth, FontSize + 10);
  }

  void moveCursor(int i) {
    cursor = constrain(cursor + i, 0, word.length());
    update();
  }

  void deleteAtCursor() {
    if (!word.equals("")) {
      word = word.substring(0, cursor-1) + word.substring(cursor);
      cursor--;
      update();
    }
  }




  //@override 
  void setText(String s) {
    Text = trim(s);
    autoSize();
  }


  void setFontSize(int s) {
    FontSize = s;
    autoSize();
  }


  //automatic resize control, so it fits the text
  private void autoSize() {
    pg.beginDraw();
    pg.textSize(FontSize);
    //Width = (int)constrain(pg.textWidth(Text), MinWidth, MaxWidth);
    Height = (int)constrain(FontSize + pg.textDescent() + pg.textAscent(), MinHeight, MaxHeight);
  }
}
